// midi.h library info http://playground.arduino.cc/Main/MIDILibrary
// MIDI_Class Class Reference http://arduinomidilib.sourceforge.net/a00001.html

void HandleNoteOn(byte channel, byte pitch, byte velocity) 
{ 
  // Do whatever you want when you receive a Note On.
  // Try to keep your callbacks short (no delays ect) as the contrary would slow down the loop()
  // and have a bad impact on real-time performance.
  if (channel == 10 && pitch < 64) // receive patch changes on channel 10; there are only 64 memory locations
  {      
    if (checkToc(pitch))
    {
      recall = true; // we're ready to recall a preset
      pageNum = pitch*7; // one memory location is 7 pages long
      cued = pitch; // the pattern that will play next is the value of "pitch" - need this to blink lights
      buttonMode = 4; // make sure we're in trigger mode, if we weren't already     
    }
  }
  else if (buttonMode != 7 && midiNoteEntry && !seqRunning && pitch > 0 && velocity != 0) // some instruments send note off as note on with 0 velocity
  {
    seqNotes[seqCurrentStep] = pitch;
    seqCurrentStep = (seqCurrentStep + 1) % 32;
    delay(2);
  }
}

void HandleClock(void) // what to do for 24ppq clock pulses
{
  seqRunning = false;
  pulse++; // 6 for a 16th
  autoCounter = (autoCounter < 191) ? autoCounter + 1 : 0; 
  static boolean long16th = true; // is this the first/longer 16th of the swing pair?
  char swingPulse = swing / 86; // to give us some swing when synced to midi clock (sadly only 2 levels: full swing and half swing)
  // takes a value between 0 and 255 and returns 0, 1 or 2
  swingPulse = (!long16th) ? (0 - swingPulse) : swingPulse;                              
  if (pulse >= (6 + swingPulse))
  {
    currentTime = millis();
    /*
    digitalWrite(LEDlatchPin, 0);
     shiftOut(LEDdataPin, LEDclockPin, B11111111); 
     shiftOut(LEDdataPin, LEDclockPin, B00000000);
     shiftOut(LEDdataPin, LEDclockPin, B00000000);
     shiftOut(LEDdataPin, LEDclockPin, B00000000);
     shiftOut(LEDdataPin, LEDclockPin, B00000000); 
     //no longer needs to listen for information
     digitalWrite(LEDlatchPin, 1);
     */
    pulse = 0;
    seqMidiStep = true;
    seqTrueStep = (seqTrueStep+1)%32; 
    if (seqLength == 32)
      seqCurrentStep = seqTrueStep;
    else
    {
      seqCurrentStep = (seqCurrentStep + 1)%seqLength;
    }
    long16th = !long16th;

    /*
    if (sixteenthStartTime != 0)
     {
     sixteenthDur = currentTime - sixteenthStartTime;
     if (seqTrueStep%2 != 0) // it's an uneven step, ie. the 2nd shorter 16th of the swing-pair
     swing16thDur = (2*sixteenthDur) - swing16thDur;
     else // it's an even step ie. the first longer 16th of the swing-pair
     {
     swing16thDur = map(swing, 0, 255, sixteenthDur, ((2*sixteenthDur)/3)*2);
     }
     onDur = (onDur > swing16thDur) ? swing16thDur : onDur;
     } 
     */
    sixteenthStartTime = millis();
  }
}

void HandleStart (void)
{
  seqTrueStep = 0;
  seqCurrentStep = 0;
  autoCounter = 0;
  sixteenthStartTime = 0;
}

void HandleStop (void)
{
  seqTrueStep = 0;
  seqCurrentStep = 0;
  autoCounter = 0;
  sixteenthStartTime = 0;
  if (midiNoteOut)
  {
    for (byte i = 0; i < 128; i++) // turn off all midi notes
      MIDI.sendNoteOff(i, 0, 1);
  }
  seqRunning = false;
  noteOn = false; 
}

void HandleSPP (unsigned int beats)
{
  if (beats == 0)
  {
    // seqTrueStep = beats%32;
    digitalWrite(LEDlatchPin, 0);
    shiftOut(LEDdataPin, LEDclockPin, B11111111); 
    shiftOut(LEDdataPin, LEDclockPin, B00000000);
    shiftOut(LEDdataPin, LEDclockPin, B00000000);
    shiftOut(LEDdataPin, LEDclockPin, B00000000);
    shiftOut(LEDdataPin, LEDclockPin, B00000000); 
    //no longer needs to listen for information
    digitalWrite(LEDlatchPin, 1);
  }

}


