// *************************************
//           HELPER FUNCTIONS
// *************************************

// some functions to streamline our code with regards to the bitwise byte arrays
void clearAll()
{
  clearSelected(); // these are our own functions, declared below
  clearActive();  
  clearTied();
  clearSlide();
  clearNotes();
  automate = false;    
  seqRunning = true;
  pulse = 0;
  seqTrueStep = 0;
  nowPlaying = 255;
  cued = 255;
}

void clearSelected()
{
  for (byte i = 0; i < 4; i++)
    seqStepSelected[i] = B00000000;
}

void clearActive()
{
  for (byte i = 0; i < 4; i++)
    seqStepMute[i] = B00000000;
}

void clearTied()
{
  for (byte i = 0; i < 4; i++)
    seqStepTied[i] = B00000000;
}

void clearSlide()
{
  for (byte i = 0; i < 4; i++)
    seqStepSlide[i] = B00000000;
}

void clearNotes()
{
  for (byte i = 0; i < 32; i++)
    seqNotes[i] = B00000000;
}

void clearJust() // clear the "just" arrays
{
  for (byte i = 0; i < 40; i++)
  {
    justpressed[i] = 0;
    justreleased[i] = 0;
  }
}
void clearEdit() // clear the "edit" arrays
{
  *edit0 = B00000000;
  *edit1 = B00000000;
  *edit2 = B00000000;
  *edit3 = B00000000;
}
boolean checkSelected(byte thisStep) // check whether a step is selected or not - returns true if a step is selected, false if not
{
  if (seqStepSelected[thisStep/8] & (1<<thisStep%8)) //seqStepSelected is an array of unsigned bytes, so dividing thisStep by 8 will return 0, 1, 2, or 3 (ie. the byte we're concerned with)
    return true;
  else
    return false; 
}

boolean checkMute(byte thisStep) // check whether a step is muted or not - returns true if a step is mute, false if not
{
  if (seqStepMute[thisStep/8] & (1<<thisStep%8)) //seqStepMute is an array of unsigned bytes, so dividing thisStep by 8 will return 0, 1, 2, or 3 (ie. the byte we're concerned with)
    return true;
  else
    return false;  
}

boolean checkTied(byte thisStep) // check whether a step is tied or not - returns true if a step is tied, false if not
{
  if (seqStepTied[thisStep/8] & (1<<thisStep%8)) //seqStepTied is an array of unsigned bytes, so dividing thisStep by 8 will return 0, 1, 2, or 3 (ie. the byte we're concerned with)
    return true;
  else
    return false; 
}

boolean checkAccent(byte thisStep) // check whether this step has an accent or not - returns true if a step is tied, false if not
{
  if (seqAccent[thisStep/8] & (1<<thisStep%8)) //seqAccent is an array of unsigned bytes, so dividing thisStep by 8 will return 0, 1, 2, or 3 (ie. the byte we're concerned with)
    return true;
  else
    return false; 
}

boolean checkSlide(byte thisStep) // check whether a step is a slide or not - returns true if it's a slide, false if not
{
  if (seqStepSlide[thisStep/8] & (1<<thisStep%8)) //seqStepSlide is an array of unsigned bytes, so dividing thisStep by 8 will return 0, 1, 2, or 3 (ie. the byte we're concerned with)
    return true;
  else
    return false; 
}

// POTS
// reading from 4051
// this function returns the analog value for the given channel
int getValue( int channel)
{
  // set the selector pins HIGH and LOW to match the binary value of channel
  for(int bit = 0; bit < 3; bit++)
  {
    int pin = select[bit]; // the pin wired to the multiplexer select bit
    int isBitSet =  bitRead(channel, bit); // true if given bit set in channel
    digitalWrite(pin, isBitSet);
  }
  //return map((constrain(analogRead(analogPin), 40, 1000)), 40, 1000, 0, 1023); // we're only using readings between 40 and 1000, as the extreme ends are unstable
  int reading = analogRead(analogPin);
  return reading;
}

// a handy DIFFERENCE FUNCTION
int difference(int i, int j)
{
  int k = i - j;
  if (k < 0)
    k = j - i;
  return k;
}

// things that need to happen as a consequence of the bpm changing
void bpmChange()
{
  clockPulseDur = (60000/bpm)/24;
  sixteenthDur = (60000/bpm)/4;
  if (onDur >= sixteenthDur && durationPU != 9999)
  {
    onDur = sixteenthDur - 10;
    offTime = sixteenthStartTime + onDur;
    // swing16thDur = map(swing, 0, 255, sixteenthDur, ((2*sixteenthDur)/3)*2);
  }      
  //swing16thDur = map(swing, 0, 255, sixteenthDur, ((2*sixteenthDur)/3)*2);
  if (seqTrueStep%2 != 0) // it's an uneven step, ie. the 2nd shorter 16th of the swing-pair
    swing16thDur = (2*sixteenthDur) - swing16thDur;
  else // it's an even step ie. the first longer 16th of the swing-pair
  {
    swing16thDur = map(swing, 0, 255, sixteenthDur, ((2*sixteenthDur)/3)*2);
  }
}

// read the table of contents (TOC) from the EEPROM (page 448)
// also recalls preferences
void tocRead()
{
  mem_1.readPage(448, rwBuffer);
  for (byte i = 0; i < 14; i++)
  {
    toc[i] = rwBuffer[i];
  }

  // remember to save preferences too, since they're on the same EEPROM page
  midiNoteEntry = rwBuffer[14]; // send notes to the sequencer from incoming MIDI data (MIDI keyboard, etc.)
  midiNoteOut = rwBuffer[15]; // send out MIDI notes from the sequencer to the next device in the chain
  midiAutomationOut = rwBuffer[16]; // send out automation data as CC16
  notify = rwBuffer[17]; // are we doing the scrolling text?
}



void tocWrite(byte addLoc)
{
  byte tocByte = addLoc/8; // in which byte of the toc array is the location (each byte corresponds to a row of buttons)
  toc[tocByte] = bitSet(toc[tocByte], addLoc%8); // turns the appropriate byte on (1)
  for (byte i = 0; i < 14; i++) // update the read/write buffer with the current toc
    rwBuffer[i] = toc[i];

  // remember to save preferences too, since they're on the same EEPROM page
  rwBuffer[14] = midiNoteEntry; // send notes to the sequencer from incoming MIDI data (MIDI keyboard, etc.)
  rwBuffer[15] = midiNoteOut; // send out MIDI notes from the sequencer to the next device in the chain
  rwBuffer[16] = midiAutomationOut; // send out automation data as CC16
  rwBuffer[17] = notify; // are we doing the scrolling text?

  mem_1.writePage(448, rwBuffer); // write the buffer to the toc page on the EEPROM (448)
}

void tocClear(byte clearLoc)
{
  byte tocByte = clearLoc/8; // in which byte of the toc array is the location (each byte corresponds to a row of buttons)
  toc[tocByte] = bitClear(toc[tocByte], clearLoc%8); // turns the appropriate byte off (0)
  for (byte i = 0; i < 14; i++) // update the read/write buffer with the current toc
    rwBuffer[i] = toc[i];

  // remember to save preferences too, since they're on the same EEPROM page
  rwBuffer[14] = midiNoteEntry; // send notes to the sequencer from incoming MIDI data (MIDI keyboard, etc.)
  rwBuffer[15] = midiNoteOut; // send out MIDI notes from the sequencer to the next device in the chain
  rwBuffer[16] = midiAutomationOut; // send out automation data as CC16
  rwBuffer[17] = notify; // are we doing the scrolling text?

  mem_1.writePage(448, rwBuffer); // write it to the toc page on the EEPROM (448)

}

boolean checkToc(byte thisLoc) // check whether a memory location is already used or not
{
  byte tocByte = thisLoc/8; // in which byte of the toc array is the location (each byte corresponds to a row of buttons)
  if bitRead(toc[tocByte], thisLoc%8)
    return true;
  else
    return false;  
}

void savePrefs()
{
  //preferences are saved on the same EEPROM page (448) as the toc, so make sure to resave the toc when saving preferences
  for (byte i = 0; i < 14; i++) // update the read/write buffer with the current toc
    rwBuffer[i] = toc[i];

  rwBuffer[14] = midiNoteEntry; // send notes to the sequencer from incoming MIDI data (MIDI keyboard, etc.)
  rwBuffer[15] = midiNoteOut; // send out MIDI notes from the sequencer to the next device in the chain
  rwBuffer[16] = midiAutomationOut; // send out automation data as CC16
  rwBuffer[17] = notify; // are we doing the scrolling text?

  mem_1.writePage(448, rwBuffer); // write the buffer to the toc page on the EEPROM (448) 
}

void savePatch()
{
  static boolean pageSaved = false;
  static byte progress; // we want to stagger the save process over consecutive loops, in case the process takes too long and causes an audible delay
  if (!pageSaved)
  {    
    clearJust();
    longPress = 0;
    //mem_1.readPage(pageNum, rwBuffer);
    pageSaved = true;
    progress = 0;
  }

  switch (progress) {
  case 0:
    // pack our 64 byte buffer with the goodies we want to send to the EEPROM
    for (byte i = 0; i < 32; i++)
    {
      rwBuffer[i] = seqNotes[i];      
    }
    for (byte i = 0; i < 4; i++)
    {
      rwBuffer[i + 32] = seqStepMute[i]; // step mutes
    }
    for (byte i = 0; i < 4; i++)
    {
      rwBuffer[i + 36] = seqStepTied[i];  // step ties
    }
    for (byte i = 0; i < 4; i++)
    {
      rwBuffer[i + 40] = seqStepSlide[i]; //step slides
    }

    rwBuffer[44] = bpm;               // bpm

      rwBuffer[45] = highByte(durationPU);   // duration (int split into a high and low byte)

    rwBuffer[46] = lowByte(durationPU);     

    rwBuffer[47] = swing; //swing

      rwBuffer[48] = seqLength; // the length of the pattern

    rwBuffer[49] = followAction; // what happens when the end of the pattern is reached?

    rwBuffer[50] = xorDistortion; // distortion

      rwBuffer[51] = voice1; // voice1 wave type

    rwBuffer[52] = voice2; // voice1 wave type

    rwBuffer[53] = voice3; // voice1 wave type

    progress = 1;
    break;

  case 1:
    if (voice2PitchPU != 9999) // if pickup isn't currently true for this variable
    {
      rwBuffer[54] = highByte(voice2PitchPU); // voice2 pitch pickup value - for transposing voice 2 relative to voice 1
      rwBuffer[55] = lowByte(voice2PitchPU); // stored as 2 bytes - the int is split into a high and low byte 
    }
    else // pickup is true, so get the value directly from the pots
    {
      int tmp = constrain(pot4, 40, 1000);
      rwBuffer[54] = highByte(tmp); // voice2 pitch pickup value - for transposing voice 2 relative to voice 1
      rwBuffer[55] = lowByte(tmp); // stored as 2 bytes - the int is split into a high and low byte 
    }
    progress = 2;
    break;

  case 2:
    if (voice3PitchPU != 9999)
    {
      rwBuffer[56] = highByte(voice3PitchPU); // voice3 pitch pickup value - for transposing voice 3 relative to voice 1
      rwBuffer[57] = lowByte(voice3PitchPU);
    }
    else // pickup is true, so get the value directly from the pots
    {
      int tmp = constrain(pot2, 40, 1000);
      rwBuffer[56] = highByte(tmp); // voice3 pitch pickup value - for transposing voice 3 relative to voice 1
      rwBuffer[57] = lowByte(tmp); // stored as 2 bytes - the int is split into a high and low byte 
    }

    rwBuffer[58] = automate; // is automation on or off

    for (byte i = 0; i < 4; i++)
    {
      rwBuffer[i + 59] = seqAccent[i]; // accents
    }

    rwBuffer[63] = accentLevel;

    progress = 3;
    break;

  case 3:
    // the first value is the number of the page (0 - 511, 64 bytes on a page)
    // the second is our EEPROM read/write buffer
    mem_1.writePage(pageNum, rwBuffer);
    progress = 4;
    break;

  case 4:
    // write automation page 1/3
    for (byte i = 0; i < 64; i++) // there will be 64 bytes
    {
      rwBuffer[i] = seqAutomate[i];
    }  
    mem_1.writePage(pageNum + 1, rwBuffer);
    progress = 5;
    break;

  case 5:
    // write automation page 2/3
    for (byte i = 0; i < 64; i++) // 64 bytes
    {
      rwBuffer[i] = seqAutomate[i + 64];
    }  
    mem_1.writePage(pageNum + 2, rwBuffer);
    progress = 6;
    break;

  case 6:
    // write automation page 3/3
    for (byte i = 0; i < 64; i++) // 64 bytes
    {
      rwBuffer[i] = seqAutomate[i + 128];
    }  
    mem_1.writePage(pageNum + 3, rwBuffer);

    save = false;
    pageSaved = false;
    break;
  }
}

void loadPatch(byte patchNumber) // load the specified location number
{
  static boolean pageLoaded = false;
  static byte progress; // we want to stagger distributing the loaded values over consecutive loops, in case the process takes too long and causes an audible delay
  saved = true; // needed for follow actions 1 and 2
  recall = false;
  pageNum = patchNumber * 4; // each location is 4 pages long
  if (!pageLoaded)
  {
    clearSelected();
    clearJust();
    mem_1.readPage(pageNum, rwBuffer);
    pageLoaded = true;
    progress = 0;
    loadContinue = true;
    cued = patchNumber; // for the duration of the load process, we set cued = patchNumber so we can loadContinue with loadPatch(cued)
  }

  switch (progress) {
  case 0:
    //if (seqCurrentStep%15 == 0)
    //{
    for (byte i = 0; i < 32; i++) // unpack the rwBuffer to the appropriate variables
    {
      seqNotes[i] = rwBuffer[i];           
    }
    progress = 1; 
    // }
    break;

  case 1:
    for (byte i = 0; i < 4; i++)
    {
      seqStepMute[i] = rwBuffer[i + 32];
    }
    progress = 2;
    break;

  case 2:
    for (byte i = 0; i < 4; i++)
    {
      seqStepTied[i] = rwBuffer[i + 36];
    }
    progress = 3;
    break;

  case 3:    
    for (byte i = 0; i < 4; i++)
    {
      seqStepSlide[i] = rwBuffer[i + 40];
    }
    progress = 4;
    break;

  case 4:
    nowPlaying = cued; // the pattern playing is the one that was cued before

    bpm = rwBuffer[44]; // recall the bpm  
    bpmPU = constrain((bpm-45)*5, 40, 1000); // calculate an approximate bpmPU, since we don't want to waste space saving it, heavy math to work out the real one 
    bpmChange(); // call our helper function to take care of everything that goes along with changing tempo/bpm

    durationPU = word(rwBuffer[45], rwBuffer[46]); // recall duration - this is the actual pickup value, so nothing to estimate
    onDur = map(durationPU, 40, 1000, 15, (sixteenthDur - 10));

    swing = rwBuffer[47]; // recall the swing value
    swingPU = constrain(swing*5, 40, 1000); // approximate a pickup value

    seqLength = rwBuffer[48]; // recall the length of the pattern
    seqLengthPU = constrain(seqLength*31, 40, 1000);

    followAction = rwBuffer[49]; // recall the follow action for this pattern

    progress = 5;
    break;

  case 5:
    xorDistortion = rwBuffer[50]; //recall the xor distortion value
    xorDistPU = constrain(xorDistPU<<2, 40, 1000);

    voice1 = rwBuffer[51];
    voice1PU = constrain(voice1<<8, 40, 1000);

    voice2 = rwBuffer[52];
    voice2PU = constrain(voice2<<8, 40, 1000);

    voice3 = rwBuffer[53];
    voice3PU = constrain(voice3<<8, 40, 1000);
    progress = 6;
    break;

  case 6:
    voice2PitchPU = word(rwBuffer[54], rwBuffer[55]);

    voice3PitchPU = word(rwBuffer[56], rwBuffer[57]);

    automate = rwBuffer[58];

    rwBuffer[58] = automate; // is automation on or off

    for (byte i = 0; i < 4; i++)
    {
      seqAccent[i] = rwBuffer[i + 59]; // accents
    }

    accentLevel = rwBuffer[63];

    progress = 7;
    break;

  case 7:
    // load automation page 1/6
    mem_1.readPage(pageNum + 1, rwBuffer);
    for (byte i = 0; i < 64; i++)
    {
      seqAutomate[i] = rwBuffer[i];
    }

    progress = 8;
    break;

  case 8:
    // load automation page 2/6
    mem_1.readPage(pageNum + 2, rwBuffer);
    for (byte i = 0; i < 64; i++)
    {
      seqAutomate[i + 64] = rwBuffer[i];
    }
    progress = 9;
    break;

  case 9:
    // load automation page 3/6
    mem_1.readPage(pageNum + 3, rwBuffer);
    for (byte i = 0; i < 64; i++)
    {
      seqAutomate[i + 128] = rwBuffer[i];
    }

    cued = 255; // out of range since 112 patterns only - we'll use 255 to see if something is cued
    pageLoaded = false;
    loadContinue = false;
    break;
  }
}

void scrollText(char myString[])
{
  byte textLength = strlen(myString);
  // read character definitions from EEPROM
  for (byte j = 0; j < 1; j++) // how many times from 0 - 9 & A - 7
  {
    for (byte m = 0; m < textLength; m++) // how many characters
    {
      int page = 508 ;
      byte c = myString[m];
      if (c < 0)       // SPACE
      {
        c = 0;
        page = 508;
      }
      else if (c < 58)  // 0 - 10
      {
        c = (c - 47)*5;
        page = 508;
      }
      else if (c < 65)  // ?
      {
        c = 55;
      }
      else if (c < 77) // A - L
      {
        c = (c - 65)*5;
        page = 509;
      }
      else if (c < 89) // M - X
      {
        c = (c - 77)*5;
        page = 510;
      }
      else
      {
        c = (c - 89)*5;
        page = 511;
      }
      mem_1.readPage(page, rwBuffer); // read the page that the character is on
      byte charRow1 =  rwBuffer[c];
      byte charRow2 =  rwBuffer[c+1];
      byte charRow3 =  rwBuffer[c+2];
      byte charRow4 =  rwBuffer[c+3];
      byte charRow5 =  rwBuffer[c+4];

      for (byte i = 0; i < 7; i++) // how many pixels wide?
        // it's only 7 because we don't need three blank pixels before the character, 2 are enough
      {
        //shift each row one pixel (bit) to the right to make space for the new collumn 
        LEDrow1 = LEDrow1 >> 1;
        LEDrow2 = LEDrow2 >> 1;
        LEDrow3 = LEDrow3 >> 1;
        LEDrow4 = LEDrow4 >> 1;
        LEDrow5 = LEDrow5 >> 1;

        if (bitRead(charRow1, 6-i)) // shorthand for "if ((bitRead(charRow1, i) == 1)"
          bitSet(LEDrow1, 7);
        if (bitRead(charRow2, 6-i)) // 
          bitSet(LEDrow2, 7);
        if (bitRead(charRow3, 6-i))
          bitSet(LEDrow3, 7);
        if (bitRead(charRow4, 6-i))
          bitSet(LEDrow4, 7);
        if (bitRead(charRow5, 6-i))
          bitSet(LEDrow5, 7);

        for (int k = 0; k < 10; k++) // how many times through the update cycle before we move to the next vertical row of pixels
        {
          updateLeds();
          delay(5);
        }
      }
    }
  }
}

void randomPattern(byte type)
{
  clearAll();
  randomSeed(millis());
  byte lastNote = tempNote; // the last note value selected in edit mode
  byte tmp = 0;
  for (byte i = 0; i < 32; i++)
  {
    tmp = random(1, 101); // NB: in the random funtion, the upper bound value is excluded, so the highest possible value here is 100 (the lower bound is included)
    if (tmp < noteDensity) // x% chance there will be a note on a step, x is a value coming from pot1 (range between 20% - 100%)
    {
      if (type == 0)
      {
        seqNotes[i] = lastNote + random(-12, 13); // this note can be any inteval from an octave up to an octave down from the last
        lastNote = seqNotes[i];
      }

      else if (type == 1) // major
      {
        tmp = random(1, 100);
        if (tmp < 80) // 80% chance the note will be one of the triad
        {  
          seqNotes[i] = lastNote + pgm_read_byte_near(majorTriad + random(0, 4));
        }
        else // 20% chance that it'll be one of the other intervals of the scale
        {
          seqNotes[i] = lastNote + pgm_read_byte_near(majorIntervals + random(0, 4));  
        }
      }

      else if (type == 2) // minor
      {
        tmp = random(1, 100);
        if (tmp < 80) // 80% chance the note will be one of the triad
        {  
          seqNotes[i] = lastNote + pgm_read_byte_near(minorTriad + random(0, 4));
        }
        else // 20% chance that it'll be one of the other intervals of the scale
        {
          //seqNotes[i] = lastNote;
          seqNotes[i] = lastNote + pgm_read_byte_near(minorIntervals + random(0, 4));  
        }
      }

      else if (type == 3) // pentatonic
      {
        seqNotes[i] = lastNote + pgm_read_byte_near(pentatonic + random(0, 6));
      }

      tmp = random(1, 100);
      if (tmp < 35) // 20% chance that this step will slide
      {
        if (i < 8)
          seqStepSlide[0] = bitSet(seqStepSlide[0],i); 
        else if (i < 16)
          seqStepSlide[1] = bitSet(seqStepSlide[1],i%8);
        else if (i < 24)
          seqStepSlide[2] = bitSet(seqStepSlide[2],i%8);
        else if (i < 32)
          seqStepSlide[3] = bitSet(seqStepSlide[3],i%8);
      } 
    }
    else
    {
      tmp = random(1, 100);
      if (tmp < 35) // there's a 35% chance that the step will be tied
      {
        if (i < 8)
          seqStepTied[0] = bitSet(seqStepTied[0],i); 
        else if (i < 16)
          seqStepTied[1] = bitSet(seqStepTied[1],i%8);
        else if (i < 24)
          seqStepTied[2] = bitSet(seqStepTied[2],i%8);
        else if (i < 32)
          seqStepTied[3] = bitSet(seqStepTied[3],i%8);
      }
    }
  }
}

void buttonCheckSelected()
{
  // check if a switch was just pressed and toggle the appropriate bit in the selected array
  for (byte i = 0; i < 32; i++)
  { 
    if (justpressed[i])
      if (i < 8)
      {    
        *edit0 ^= (1<<i); // toggle the bit, ie. turn 1 to 0, and 0 to 1
        if (buttonMode == 0)
          seqStepMute[0] = bitClear(seqStepMute[0], i); // turn mute off - don't toggle
        else if (buttonMode == 1 && seqNotes[i] == 0) // we're in mute mode and the step is not active
          seqStepMute[0] = bitClear(seqStepMute[0], i); // turn mute off - don't toggle
        else if (buttonMode == 3 && seqNotes[i] == 0) // we're in slide mode and the step is not active
          seqStepSlide[0] = bitClear(seqStepSlide[0], i); // turn slide off - don't toggle
      }
      else if (i < 16)
      {
        *edit1 ^= (1<<(i%8));
        if (buttonMode == 0)
          seqStepMute[1] = bitClear(seqStepMute[1], i%8); // turn it off - don't toggle
        else if (buttonMode == 1 && seqNotes[i] == 0) // turn mute off if the step is not active
          seqStepMute[1] = bitClear(seqStepMute[1], i%8); // turn mute off - don't toggle
        else if (buttonMode == 3 && seqNotes[i] == 0) // we're in slide mode and the step is not active
          seqStepSlide[1] = bitClear(seqStepSlide[1], i%8); // turn slide off - don't toggle
      }
      else if (i < 24)
      {
        *edit2 ^= (1<<(i%8));
        if (buttonMode == 0)
          seqStepMute[2] = bitClear(seqStepMute[2], i%8); // turn it off - don't toggle
        else if (buttonMode == 1 && seqNotes[i] == 0) // turn mute off if the step is not active
          seqStepMute[2] = bitClear(seqStepMute[2], i%8); // turn mute off - don't toggle
        else if (buttonMode == 3 && seqNotes[i] == 0) // we're in slide mode and the step is not active
          seqStepSlide[2] = bitClear(seqStepSlide[2], i%8); // turn slide off - don't toggle
      }
      else if (i < 32)
      {
        *edit3 ^= (1<<(i%8));
        if (buttonMode == 0)
          seqStepMute[3] = bitClear(seqStepMute[3], i%8); // turn it off - don't toggle
        else if (buttonMode == 1 && seqNotes[i] == 0) // turn mute off if the step is not active
          seqStepMute[3] = bitClear(seqStepMute[3], i%8); // turn mute off - don't toggle
        else if (buttonMode == 3 && seqNotes[i] == 0) // we're in slide mode and the step is not active
          seqStepSlide[3] = bitClear(seqStepSlide[3], i%8); // turn slide off - don't toggle
      }  
  } 
}

void tapTempo()
{
  static byte i = 0; // the current index in the bpmTaps array
  static unsigned long lastTap = 0;  
  static byte tapsNumber = 1; // how many bpms have been added to the bpmTaps array
  unsigned long now = millis(); // the current time in milliseconds

    if (lastTap == 0)
  {
    lastTap = now;
    pulse = 0;
    seqTrueStep = 0;
  }
  else if ((now - lastTap) < 1334) // a quarter is 1333ms long at 45bpm (our slowest allowed bpm)
  {
    bpmTaps[i] = 60000/(now - lastTap);
    i = (i + 1) % 5; 
    if (tapsNumber > 1)
    {
      unsigned int bpmTotal = 0;  
      for (byte j = 0; j < (tapsNumber); j++)
      {
        bpmTotal += bpmTaps[j];
      }
      bpm = bpmTotal / (tapsNumber); //get the average bpm
    }
    else
    {
      bpm = bpmTaps[0];
    }
    lastTap = now;
    bpmPU = constrain((bpm-45)*5, 40, 1000); // calculate an approximate bpmPU, since we don't want to waste space saving it, heavy math to work out the real one
    bpmChange();
    pulse = 0;
    seqTrueStep = 0;
    tapsNumber = (tapsNumber < 5) ? tapsNumber + 1 : 5; // we want to find the average of at most all 5 bpms in the array
  }
  else 
  {
    i = 0;
    tapsNumber = 1;
    lastTap = now;
    pulse = 0;
    seqTrueStep = 0;
  }
}

/**********************************************************************************************
 ***
 *** ShiftOut to address 42 LEDs via 4 74HC595s serial to parallel shifting registers 
 *** Based on http://arduino.cc/en/Tutorial/ShiftOut
 ***
 ***********************************************************************************************/
//***TOP*OF*THE*SKETCH**************************************************************************
// remember to add the following to the start of your sketch before (before setup)
/*
int LEDlatchPin = 2;
 int LEDclockPin = 3;
 int LEDdataPin = 4;
 */
//***SETUP**************************************************************************************
// and add the following pin modes to setup()
/*
pinMode(LEDlatchPin, OUTPUT);
 pinMode(LEDclockPin, OUTPUT); 
 pinMode(LEDdataPin, OUTPUT);
 */

void shiftOut(int myLEDdataPin, int myLEDclockPin, byte myDataOut) {
  // This shifts 8 bits out MSB first, 
  // on the rising edge of the clock,
  // clock idles low

  //internal function setup
  int i=0;
  int pinState;
  pinMode(myLEDclockPin, OUTPUT);
  pinMode(myLEDdataPin, OUTPUT);

  //clear everything out just in case to
  //prepare shift register for bit shifting
  digitalWrite(myLEDdataPin, 0);
  digitalWrite(myLEDclockPin, 0);

  //for each bit in the byte myDataOut
  //NOTICE THAT WE ARE COUNTING DOWN in our for loop
  //This means that %00000001 or "1" will go through such
  //that it will be pin Q0 that lights. 
  for (i=7; i>=0; i--)  {
    digitalWrite(myLEDclockPin, 0);

    //if the value passed to myDataOut and a bitmask result 
    // true then... so if we are at i=6 and our value is
    // %11010100 it would the code compares it to %01000000 
    // and proceeds to set pinState to 1.
    if ( myDataOut & (1<<i) ) {
      pinState= 1;
    }
    else {	
      pinState= 0;
    }

    //Sets the pin to HIGH or LOW depending on pinState
    digitalWrite(myLEDdataPin, pinState);
    //register shifts bits on upstroke of clock pin  
    digitalWrite(myLEDclockPin, 1);
    //zero the data pin after shift to prevent bleed through
    digitalWrite(myLEDdataPin, 0);
  }

  //stop shifting
  digitalWrite(myLEDclockPin, 0);
}

// this function shifts out the the 4 bytes corresponding to the led rows
void updateLeds(void)
{
  static boolean lastSentTop = false; // we want to alternate sending the top 2 and bottom 3 rows to prevent an edge case where 4 rows of LEDs lit at the same time sourcing too much current
  //ground LEDlatchPin and hold low for as long as you are transmitting
  digitalWrite(LEDlatchPin, 0);
  if (!lastSentTop) // send the top to rows
  {
    shiftOut(LEDdataPin, LEDclockPin, B00000000);
    shiftOut(LEDdataPin, LEDclockPin, B00000000); 
    shiftOut(LEDdataPin, LEDclockPin, B00000000);
    shiftOut(LEDdataPin, LEDclockPin, LEDrow2); 
    shiftOut(LEDdataPin, LEDclockPin, LEDrow1);
    lastSentTop = true;
  }
  else // ie. lastSentTop is true, then send the bottom 3 rows
  {
    shiftOut(LEDdataPin, LEDclockPin, LEDrow5);
    shiftOut(LEDdataPin, LEDclockPin, LEDrow4); 
    shiftOut(LEDdataPin, LEDclockPin, LEDrow3);
    shiftOut(LEDdataPin, LEDclockPin, B00000000); 
    shiftOut(LEDdataPin, LEDclockPin, B00000000);
    lastSentTop = false;
  }
  //return the latch pin high to signal chip that it 
  //no longer needs to listen for information
  digitalWrite(LEDlatchPin, 1);
}














































